# Copyright (c) 2016-2023 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from abc import ABC
from typing import Optional, cast

from astropy.time import Time

from lucupy.minimodel.observationmode import ObservationModes
from lucupy.minimodel.resource import Resource, Resources
from lucupy.minimodel.wavelength import Wavelengths

__all__ = [
    'ObservatoryProperties',
]


class ObservatoryProperties(ABC):
    """Observatory-specific methods.

       These are not tied to other components or
       structures, and allow computations to be implemented in one place.

    """
    _properties: Optional[ObservatoryProperties] = None

    @staticmethod
    def set_properties(cls) -> None:
        """Set properties for an specific Observatory

        Raises:
            ValueError: Illegal properties value.

        """
        if not issubclass(cls, ObservatoryProperties):
            raise ValueError('Illegal properties value.')
        ObservatoryProperties._properties = cls()

    @staticmethod
    def _check() -> ObservatoryProperties:
        """ Check if the properties have been set.

        Raises:
            ValueError: Properties have not been set.
        """
        if ObservatoryProperties._properties is None:
            raise ValueError('Observatory properties have not been set.')
        return cast(ObservatoryProperties, ObservatoryProperties._properties)

    @staticmethod
    def determine_standard_time(resources: Resources,
                                wavelengths: Wavelengths,
                                modes: ObservationModes,
                                cal_length: int) -> Time:
        """Determine standard time for a specific Observatory

        Args:
            resources (Resources): Set of Resources(instruments, mask, etc).
            wavelengths (Wavelengths): An array of Wavelengths to be observed.
            modes (ObservationModes): The different modes of observation.
            cal_length (int): The length (in seconds) of a calibration.

        Returns:
            Time: Value(s) of standard time
        """
        return ObservatoryProperties._check().determine_standard_time(
            resources,
            wavelengths,
            modes,
            cal_length
        )

    @staticmethod
    def nir_instruments() -> Resources:
        return ObservatoryProperties._check().nir_instruments()

    @staticmethod
    def instruments() -> Resources:
        return ObservatoryProperties._check().instruments()

    @staticmethod
    def is_nir_instrument(resource: Resource) -> bool:
        return ObservatoryProperties._check().is_nir_instrument(resource)

    @staticmethod
    def is_instrument(resource: Resource) -> bool:
        """Determine if the given resource is an instrument or not.

        Args:
            resource (Resource): An instrument.

        Returns:
            bool: True is the resource is an instrument of the Observatory, otherwise False.
        """
        return ObservatoryProperties._check().is_instrument(resource)



# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from dataclasses import dataclass, field, InitVar
from typing import FrozenSet, Optional, TypeAlias, final

from .resource_type import ResourceType
from lucupy.decorators import immutable

__all__ = [
    'Resource',
    'Resources',
]


@final
@immutable
@dataclass(frozen=True)
class Resource:
    """This is a general observatory resource.
    It can consist of

    * a guider;
    * an instrument;
    * a part of an instrument;
    * an observatory (Site); or even
    * personnel

    and is used to determine what observations can be
    performed at a given time based on the resource availability.

    ***NOTE:***
    Instances of Resource should NEVER be created directly through this dataclass.
    Resources are flyweight objects, and should ONLY be created through ResourceManager.lookup_resource.

    Attributes:
        id (str): Resource id or name.
        description (str): Short description.
    """
    id: str
    description: Optional[str] = None
    type: Optional[ResourceType] = ResourceType.NONE
    legal_creation: InitVar[bool] = field(default=False)

    def __post_init__(self, legal_creation: bool) -> None:
        if not legal_creation:
            raise RuntimeError(f'Resource object {self.id} attempted to be created directly. '
                               'All resources must be accessed through ResourceManager.lookup_resource.')
        if self.id is None or 'NONE' in self.id.upper():
            raise ValueError('Should not have any Resources equal to None or containing "None"')

    def __hash__(self):
        """
        We only want to hash the ID. Resources with identical IDs should be considered identical.
        """
        return hash(self.id)

    def __eq__(self, other):
        """
        Resources are the same if the have the same ID.
        """
        return isinstance(other, Resource) and self.id == other.id

    def __repr__(self):
        return f"Resource(id='{self.id}')"


Resources: TypeAlias = FrozenSet[Resource]



# Copyright (c) 2016-2024 Association of Universities for Research in Astronomy, Inc. (AURA)
# For license information see LICENSE or https://opensource.org/licenses/BSD-3-Clause

from __future__ import annotations

from enum import Enum, EnumMeta
from typing import final

import astropy.units as u
from astropy.time import Time

from lucupy.minimodel.observationmode import ObservationMode, ObservationModes
from lucupy.minimodel.resource import Resource, Resources
from lucupy.minimodel.resource_type import ResourceType
from lucupy.minimodel.wavelength import Wavelengths
from lucupy.resource_manager.resource_manager import ResourceManager
from lucupy.observatory.abstract import ObservatoryProperties

__all__ = ['GeminiProperties']

rm = ResourceManager()


@final
class GeminiProperties(ObservatoryProperties):
    """
    Implementation of ObservatoryCalculations specific to Gemini.
    """

    class _InstrumentsMeta(EnumMeta):
        """Metaclass for the Instruments Class below.
        """
        def __contains__(cls, r: Resource) -> bool:  # type: ignore[override]
            return any(inst.value.id in r.id for inst in cls.__members__.values())  # type: ignore[var-annotated]

    class Instruments(Enum, metaclass=_InstrumentsMeta):
        """ Gemini-specific instruments.
        """
        FLAMINGOS2 = rm.lookup_resource(rid='Flamingos2', rtype=ResourceType.INSTRUMENT)
        NIFS = rm.lookup_resource(rid='NIFS', rtype=ResourceType.INSTRUMENT)
        NIRI = rm.lookup_resource(rid='NIRI', rtype=ResourceType.INSTRUMENT)
        IGRINS = rm.lookup_resource(rid='IGRINS', rtype=ResourceType.INSTRUMENT)
        GMOS_N = rm.lookup_resource(rid='GMOS-N', rtype=ResourceType.INSTRUMENT)
        GMOS_S = rm.lookup_resource(rid='GMOS-S', rtype=ResourceType.INSTRUMENT)
        GNIRS = rm.lookup_resource(rid='GNIRS', rtype=ResourceType.INSTRUMENT)
        GPI = rm.lookup_resource(rid='GPI', rtype=ResourceType.INSTRUMENT)
        GSAOI = rm.lookup_resource(rid='GSAOI', rtype=ResourceType.INSTRUMENT)
        PHOENIX = rm.lookup_resource(rid='Phoenix', rtype=ResourceType.INSTRUMENT)

    _STANDARD_INSTRUMENTS = frozenset({Instruments.FLAMINGOS2,
                                       Instruments.GNIRS,
                                       Instruments.NIFS,
                                       Instruments.IGRINS})

    _NIR_INSTRUMENTS: Resources = frozenset({Instruments.FLAMINGOS2,
                                             Instruments.GNIRS,
                                             Instruments.NIRI,
                                             Instruments.NIFS,
                                             Instruments.PHOENIX,
                                             Instruments.IGRINS})

    """ List: Instruments for which there are set standards.
    """
    @staticmethod
    def determine_standard_time(resources: Resources,
                                wavelengths: Wavelengths,
                                modes: ObservationModes,
                                cal_length: int) -> Time:
        """Determine the standard star time required for Gemini.

        Args:
            resources (Resources): Instruments to be used.
            wavelengths (Wavelength): Wavelengths to be observed.
            modes (ObservationModes): Observation modes.
            cal_length (int): The specific length of a calibration.

        Returns:
            Time: _description_

        Todo:
            We may only want to include specific resources, in which case, modify
            Instruments above to be StandardInstruments.

        """
        if cal_length > 1:
            # Check to see if any of the resources are instruments.
            if any(resource in GeminiProperties._STANDARD_INSTRUMENTS for resource in resources):
                if all(wavelength <= 2.5 for wavelength in wavelengths):
                    return 1.5 * u.h
                else:
                    return 1.0 * u.h
            if ObservationMode.IMAGING in modes:
                return 2.0 * u.h
            return 0.0 * u.h

    @staticmethod
    def instruments() -> Resources:
        return GeminiProperties._STANDARD_INSTRUMENTS | GeminiProperties._NIR_INSTRUMENTS

    @staticmethod
    def nir_instruments() -> Resources:
        return GeminiProperties._NIR_INSTRUMENTS

    @staticmethod
    def is_nir_instrument(resource: Resource) -> bool:
        """
        Checks in the specified Resource is a NIR Gemini Instrument.
        """
        return resource in GeminiProperties._NIR_INSTRUMENTS

    @staticmethod
    def is_instrument(resource: Resource) -> bool:
        """Checks if the resource is a Gemini instrument.

        Args:
            resource (Resource): A resource to be checked.

        Returns:
            bool: True if resource is a Gemini instrument.
        """
        return resource in GeminiProperties._STANDARD_INSTRUMENTS | GeminiProperties._NIR_INSTRUMENTS
